# info.py

APPNAME = "KmerExploR"
SHORTDESC = "KmerExploR provides information on RNA-sequencing datasets."
LICENCE = "GPL3"
VERSION = "1.1.2"
AUTHOR = "Benoit Guibert"
AUTHOR_EMAIL = "benoit.guibert@free.fr"



import sys
import os
import argparse
from shutil import copyfile
import gzip


import info
from common import *

APPPATH = os.path.dirname(os.path.realpath(__file__))
DEFAULT_TAGSET = "human-quality"


def handle_args(args):
    ### check some args
    check_args(args)
    ### check special options
    if args.dump_config: _dump_config(args)
    if args.show_tags: _show_tags(args)
    if args.list_tagsets: _list_tagsets(args)


def check_args(args):
    special_option = any([args.show_tags, args.dump_config, args.list_tagsets])
    ### --paired OR --single
    if not special_option:
        if len([True for i in (args.paired, args.single) if i]) != 1:
            sys.exit("Syntax error: one of the arguments -s/--single -p/--paired is required")
        ### At least one fastq must be define
        if not args.files:
            sys.exit("Syntax error: the following arguments are required: <files1>")
    ### output directory must be writable
    if not os.access(os.path.dirname(os.path.abspath(args.output)), os.W_OK):
        sys.exit(f"Permission denied: {os.path.dirname(os.path.abspath(args.output))}")
    ### Define user tagset tsv/fasta, config yaml, desc md and put them in args
    user_args_len = len([_ for _ in (args.config, args.tags) if _])
    if user_args_len == 2:
        ### check file presences
        if not os.path.isfile(args.tags):
            sys.exit(f"File error: {args.tags!r} not found")
        ### Is user has define a markdown description file ?
        user_desc_file = f"{os.path.splitext(args.tags.split('gz')[0])[0]}.md"
        if not os.path.isfile(user_desc_file):
            user_desc_file = None
        ### Re-define set of tag files
        args.setfiles = {'tags':args.tags, 'config': args.config, 'desc': user_desc_file}
        return
    elif user_args_len == 1 and not special_option:
        sys.exit("SyntaxError: '-C/--config' and '-T/--tags' work together")
    args.setfiles = _get_tagsets(args)


def _get_tagsets(args=None):
    basedir, subdirs, files = next(os.walk(os.path.join(APPPATH, 'tagsets')))
    tagsets = {}
    for file in files:
        if file.endswith(('tsv.gz', 'fa.gz', 'fasta.gz')):
            tagset = ('.'.join(file.split('.')[:-2]))
            ext = '.'.join(file.split('.')[-2:])
            tagsets[tagset] = ext
        elif file.endswith(('tsv', 'fa', 'fasta')):
            tagset = ('.'.join(file.split('.')[:-1]))
            ext = '.'.join(file.split('.')[-1:])
            tagsets[tagset] = ext

    ### return list of kmer sets
    if not args:
        return list(tagsets.keys())

    ### Exit if kmer set files not found
    exit = lambda tagset, tagsets : sys.exit("kmer set '{}' not found, kmer sets available:\n - {}\n"
             "Default is 'human-quality'. You can also use your own kmer set with '--tags' option."
             .format(tagset, '\n - '.join(tagsets)))

    ### find and return kmer set, yaml config and description file
    tagset = args.builtin_tags
    ext = tagsets[tagset]
    def get_setfiles(tagset, ext):
            tag_file = os.path.join(basedir, f"{tagset}.{ext}")
            config_file = os.path.join(basedir, f"{tagset}.yaml")
            desc_file = os.path.join(basedir, f"{tagset}.md")
            if not os.path.isfile(tag_file) or not os.path.isfile(config_file):
                exit(tagset, tagsets)
            if not os.path.isfile(desc_file):
                desc_file = None
            setfiles = {'tags':tag_file, 'config': config_file, 'desc': desc_file, 'ext': ext, 'tags_format': ext.rstrip('.gz')}
            return setfiles

    ### get chosen tagset
    setfiles = get_setfiles(tagset, ext)
    return setfiles



def _dump_config(args):
    """
    Copy builtin config yaml file in current directory
    """
    try:
        copyfile(args.setfiles['config'], args.dump_config)
    except:
        sys.exit(f"FileNotFoundError: file {args.dump_config!r} not found")
    sys.exit(f"Configuration dumped in file {args.dump_config!r} succesfully.")


def _show_tags(args):
    """
    show details of specified tagset (default: human-quality)
    """
    categories = {}
    ### Define tag file.
    tags_file = args.tags or args.setfiles['tags']

    ### is tag file gzipped ?
    is_gz = _is_gz_file(tags_file)
    ### is file formated as fasta or tsv ?
    tag_format = _get_file_format(tags_file, is_gz)
    ### open tag file
    try:
        if is_gz:
            fh = gzip.open(tags_file, 'rt')
        else:
            fh = open(tags_file)
    except FileNotFoundError as err:
        sys.exit(f"Error: file {tags_file!r} not found")
    ### Extract categories and predictors
    if tag_format == 'tsv':
        for line in fh:
            try:
                category, predictor = line.split('\t')[1].split('-')[:2]
                categories.setdefault(category, set()).add(predictor)
            except IndexError:
                sys.exit(f"Error: file '{tags_file!r}' malformated (show format of tag file).")
    if tag_format == 'fasta':
        for line in fh:
            try:
                if line.startswith('>'):
                    category, predictor = line[1:].rstrip().split('-')[:2]
                    categories.setdefault(category, set()).add(predictor)
            except IndexError:
                sys.exit(f"Error: file '{tags_file!r}' malformated (show format of tag file).")



    ### Display categories and predictors
    for categ,predictors in categories.items():
        print(categ)
        for predictor in predictors:
            print(f"  {predictor}")

    sys.exit()


def _list_tagsets(args):
    """
    List builtin tag sets
    """
    print("Buitin tag set (to use with -b/--builtin-tags):")
    print(" - {}\nDefault: human-quality".format('\n - '.join(_get_tagsets())))
    sys.exit()


def _is_gz_file(file):
    with open(file, 'rb') as test_f:
        return test_f.read(2) == b'\x1f\x8b'


def _get_file_format(file, gz=False):
    ### Is tag file a fasta or a tsv
    if gz:
        with gzip.open(file, 'rt') as fh:
            format = 'fasta' if fh.readline().startswith('>') else 'tsv'
    else:
        with open(file) as fh:
            format = 'fasta' if fh.readline().startswith('>') else 'tsv'
    return format



def usage():
    """
    Help function with argument parser.
    """

    ### Text at the end (command examples)
    epilog  = ("Examples:\n"
            "\n # Mandatory: -p for paired-end or -s for single:\n"
            " %(prog)s -p path/to/*.fastq.gz\n"
            "\n # -c for multithreading, -k to keep counts (input must be fastq):\n"
            " %(prog)s -p -c 16 -k path/to/*.fastq.gz\n"
            "\n # You can skip the counting step thanks to countTags output (see -k option):\n"
            " %(prog)s -p path/to/countTags/files/*.tsv\n"
            "\n # -o to choose your directory output (directory will be created),"
            "\n # --title to show title in results:\n"
            " %(prog)s -p -o output_dir --title 'Title displayed on the html page' dir/*.fastq.gz'\n"
            "\n # Advanced: use your own tag file and config.yaml file:\n"
            " %(prog)s -p --tags my_tags.tsv --config my_config.yaml dir/*.fast.gz\n"
    )
    ### Argparse
    parser = argparse.ArgumentParser(epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        # formatter_class=argparse.RawTextHelpFormatter,
    )
    # ~ method = parser.add_mutually_exclusive_group(required=True) # method = paired or single
    advanced_group = parser.add_argument_group(title='advanced features')
    special_group = parser.add_argument_group(title='extra features')
    parser.add_argument('files',
                        help='fastq or fastq.gz or tsv countTags output files.',
                        nargs='*',
                        metavar=('<file1> ...'),
    )
    parser.add_argument('-s', '--single',
                        action='store_true',
                        help='when samples are single.',
    )
    parser.add_argument('-p', '--paired',
                        action='store_true',
                        help='when samples are paired.',
    )
    parser.add_argument('-k', '--kmer-size',
                        type=int,
                        help='kmer size (default 31).',
                        default=31,
                        metavar="<int>",
    )
    parser.add_argument('-K', '--keep-counts',
                        action='store_true',
                        help='keep countTags outputs.',
    )
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help='debug.',
    )
    parser.add_argument('-b', '--builtin-tags',
                        help=f"Choose a kmer set between {_get_tagsets()} (default: human-quality)",
                        default='human-quality',
    )
    parser.add_argument('-o', '--output',
                        default=f'./{info.APPNAME.lower()}-results',
                        help=f'output directory (default: "./{info.APPNAME.lower()}-results").',
                        metavar='<output_dir>',
    )
    parser.add_argument('-l', '--list-tagsets',
                        action='store_true',
                        help="List available kmer sets",
    )
    parser.add_argument('--tmp-dir',
                        default='/tmp',
                        help='temporary files directory.',
                        metavar='<tmp_dir>',
    )
    advanced_group.add_argument('-C', '--config',
                        help="alternate config yaml file. Used with '--tags' option",
                        metavar='<file_name>',
    )
    advanced_group.add_argument('-T', '--tags',
                        help=("alternate tag file. Could be fasta or tsv file (gzip or not)."
                              " Needs '--config' option"),
                        metavar='<tag_file>',
    )
    ### Not implemented yet: hidden
    advanced_group.add_argument('-A', '--add-tags',
                        help=argparse.SUPPRESS, # 'additional tag file.',
                        metavar='<tag_file>',
    )
    special_group.add_argument('--dump-config',
                        default=None,
                        metavar='file_name',
                        help=('dump builtin config file as specified name '
                              'as yaml format and exit.'),
    )
    special_group.add_argument('--show-tags',
                        action='store_true',
                        help='print builtin categories and predictors and exit.',
    )
    parser.add_argument('--title',
                        default='',
                        help='title to be displayed in the html page.',
                        metavar="<string>"
    )
    parser.add_argument('-y', '--yes', '--assume-yes',
                        action='store_true',
                        help='assume yes to all prompt answers.',
    )
    parser.add_argument('-c', '--cores',
                        default=1,
                        type=int,
                        help='specify the number of files which can be processed simultaneously' +
                        ' by countTags. (default: 1). Valid when inputs are fastq files.',
                        metavar=('<int>'),
    )
    parser.add_argument('-v', '--version',
                        action='version',
                        version='%(prog)s version: {}'.format(info.VERSION)
    )
    ### Go to "usage()" without arguments or stdin
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    args = parser.parse_args()
    ### check, add, and modify some parameters
    handle_args(args)

    return args



#!/usr/bin/env python3

"""
Build an objet with counts.
"""


import os
import sys
import gzip
import math
from statistics import mean, median


def main():
    """ You can use this module stand alone. """

    ### Define some variables, they must be in tsv tags file passed as input
    sample = 'A'
    seq_type = 'Orientation'
    seq_id = 'VPS29'

    ### You may create args
    class args:
        method = 'mean'
        single = False
        scale = 10
    counts_files = sys.argv[1:] # files are countTags output
    counts = Counts(counts_files, args)

    print('SAMPLES (AS SINGLE):', counts.tags['samples']['single'])
    print('SAMPLES (AS PAIRED):', counts.tags['samples']['paired'])
    print('TYPES:', *counts.types)
    print('META:', counts.tags['meta'])
    print("MODE:", counts.mode)
    print(f"----------------\n get_by_sample({sample}, {seq_type}):\n", counts.get_by_sample(sample, seq_type))
    print(f"----------------\n get_counts_by_sample('{seq_type}'):\n", counts.get_counts_by_sample(seq_type))
    print(f"----------------\n get_by_category('{seq_type}'):\n", *counts.get_by_category(seq_type))
    print(f"----------------\n get_by_category in single mode('{seq_type}'):\n", *counts.get_by_category(seq_type, mode='single'))
    print('----------------\n')
    for k, v in counts.tags['counts'][seq_type][seq_id].items():
        print(seq_type, seq_id, k, v)


class Counts:
    """
    Create a dictionay of tags
    """

    def __init__(self, counts_files, args):
        """ Class initialiser """
        self.tags = {'samples': {}, 'counts': {}, 'meta': {}}
        self.counts_files = counts_files
        self.args = args
        self.mode = 'single' if args.single else 'paired'
        ### Check counts files
        # self._check_count_files()
        ### Add entry for samples
        self._get_samples_names()
        ### merge tags
        self.tags = self._do_merge_counts(args)
        ### compute mean of counts
        self._set_counts_mean()
        ### if paired samples, add dict entry with sum for foward and reverse
        if self.mode == 'paired':
            self._match_paired()
        ### Check if tags dict is consistant
        # self._check_merge()
        ### minimize counts
        self._minimize(args)
        ### get extra info from countTags summary files
        self.get_summary_info()
        ### some attributes to easier access
        self.samples = [sample for sample in self.tags['samples'][self.mode]]
        self.fastq = [sample for sample in self.tags['samples']['single']]
        self.types = [seq_type for seq_type in self.tags['counts']]
        self.meta = self.tags['meta']


    def _do_merge_counts(self, args):
        """ Function doc """
        tags            = self.tags
        pos             = 0         # position of first sample
        n_sample        = 0         # number of sample
        for index, file in enumerate(self.counts_files):
            if file[-2:].lower() == 'gz':
                fh = gzip.open(file, 'tr')
            else:
                fh = open(file)
            ### Add data
            for row in fh:
                ### Avoid first line in counts and define counts pos
                if row[0:4] == 'tag\t':
                    pos += n_sample
                    n_sample = len(row.split()[2:])
                    continue
                ### split line
                row = row.rstrip().split('\t')
                ### The kmer can be representative of several tag names (seq-type + seq-id + queue)
                tagnames = row[1].split(',')
                for tagname in tagnames:
                    try:
                        seq_type, seq_id = tagname.split('-')[:2]
                        seq_counts = row[2:]
                    except IndexError:
                        sys.exit(f"\n IndexError: missing at least one field in '{file}' file (2 fields expected, hyphen separated).")
                    ### Add sequence type if not exists (ex: sex, orientation)
                    if not seq_type in tags['counts']:
                        tags['counts'][seq_type] = {}
                    ### Add seq_id (often gene) if not exist (ex: VPS29, HIST1H2BF)
                    if not seq_id in tags['counts'][seq_type]:
                        tags['counts'][seq_type][seq_id] = {'single':[0 for _ in range(len(self.tags['samples']['single']))]}
                        tags['counts'][seq_type][seq_id]['number_counts'] = 0       # number of counts per seq_id
                    ### Add count value
                    for ii, count in enumerate(seq_counts):
                        ### Store count
                        tags['counts'][seq_type][seq_id]['single'][ii+pos] += float(count)
                    ### Store number of counts for this seq_id
                    if pos == 0:
                        tags['counts'][seq_type][seq_id]['number_counts'] += 1
            fh.close()

        ### if first sample, add rows counts in metadata, verify if same count for others samples
        nrows = 0
        for idx,seq_type in enumerate(tags['counts']):
            for seq_id in tags['counts'][seq_type]:
                nrows += len(tags['counts'][seq_type][seq_id])
        if 'nrow' not in tags['meta']:
            tags['meta']['nrow'] = nrows
        else:
            if nrows != tags['meta']['nrow']:
                print("\n InternalError: rows count should be the same for each sample.")
                print(" {} found for {} ({} expected).\n".format(nrows, tags['samples']['single'][idx] , tags['meta']['nrow']))
                sys.exit()
        return tags


    def get_summary_info(self):
        """ Function doc """
        self.tags['meta']['fastq_files'] = {}
        self.tags['meta']['total_kmers'] = 0
        self.tags['meta']['total_reads'] = 0
        summary_files = []
        ### localize summary files
        for file in self.counts_files:
            file = file.split('.')
            if '.'.join(file[-2:]) == 'tsv.gz':
                summary_files.append(f"{'.'.join(file[:-2])}.summary")
            else:
                summary_files.append(f"{'.'.join(file[:-1])}.summary")
        ### Add reads and kmers numbers for each sample in tags.meta
        for file in summary_files:
            if os.path.exists(file):
                fastq = os.path.basename('.'.join(file.split('.')[:-1]))
                self.tags['meta']['fastq_files'][fastq] = ['', '']
                with open(file) as fh:
                    for row in fh:
                        if row.startswith('total_factors'):
                            nkmers = row.split()[1]
                            self.tags['meta']['fastq_files'][fastq][0] = nkmers
                            self.tags['meta']['total_kmers'] += int(nkmers)
                        if row.startswith('total_reads'):
                            nreads = row.split()[1]
                            self.tags['meta']['fastq_files'][fastq][1] = nreads
                            self.tags['meta']['total_reads'] += int(nreads)


    def _set_counts_mean(self):
        """ Compute average for each count """
        for seq_type in self.tags['counts']:
            for seq_id in self.tags['counts'][seq_type]:
                counts = self.tags['counts'][seq_type][seq_id]['single']
                ncounts = self.tags['counts'][seq_type][seq_id]['number_counts']
                for i,count in enumerate(counts):
                    self.tags['counts'][seq_type][seq_id]['single'][i] = count / ncounts


    def _match_paired(self):
        """ Sum sample and his mate and add into dict """
        for seq_type in self.tags['counts']:
            for seq_id in self.tags['counts'][seq_type]:
                #for seq_def in self.tags['counts'][seq_type][seq_id]:
                counts = self.tags['counts'][seq_type][seq_id]['single']
                for i in range(0, len(counts), 2):
                    if not 'paired' in self.tags['counts'][seq_type][seq_id]:
                        self.tags['counts'][seq_type][seq_id]['paired'] = []
                    self.tags['counts'][seq_type][seq_id]['paired'].append(counts[i]+counts[i+1])


    def _minimize(self, args):
        """ Minimize counts """
        precision=4
        for seq_type in self.tags['counts']:
            for seq_id in self.tags['counts'][seq_type]:
                for mode, counts in self.tags['counts'][seq_type][seq_id].items():
                    if mode == 'single' or mode == 'paired':
                        for i,n in enumerate(counts):
                            if n:
                                mod = 0 if n < 1 else 1
                                p = math.pow(10,int(-math.log10(n))+precision-mod)
                                self.tags['counts'][seq_type][seq_id][mode][i] = round(n*p)/p


    def _get_samples_names(self):
        """ get counts for a sample """
        self.tags['samples']['single'] = []
        self.tags['samples']['paired'] = []
        ### Always define sample name 'single' (Required for Orientation category)
        for index, file in enumerate(self.counts_files):
            ### pick column names in the first line ('tag', 'tag_names', 'col1', 'col...', )
            if file[-2:].lower() == 'gz':
                fh = gzip.open(file, 'rt')
            else:
                fh = open(file)
            sample_names = fh.readline().rstrip().split('\t')[2:]
            for sample_name in sample_names:
                self._append_samples_names_to_count_dict(sample_name)
            fh.close
        ### If paired, add merge column names
        if self.mode == 'paired':
            paired_names = [name[:-2] for i, name in enumerate(self.tags['samples']['single']) if i % 2 == 0]
            self.tags['samples']['paired'] = paired_names


    def _append_samples_names_to_count_dict(self, sample_name):
        """ append sample name to counts dict """
        sample_name = os.path.basename(sample_name).rstrip('.fastq.gz').rstrip('.fq.gz').rstrip('.fastq')
        ### Add to samples list
        if not sample_name in self.tags['samples']['single']:
            self.tags['samples']['single'].append(sample_name)


    def _check_count_files(self):
        """ Control if files produced by countTags have same size """
        file_sizes = set()
        for file in self.counts_files:
            file_sizes.add(os.popen('wc -l < ' + file).readline())

        if len(file_sizes) > 1:
            print('\n Error : one or more of countTags output have not the same size.')
            print(" Reload command with '-d' option and check countTags outputs.\n")
            sys.exit()

        if not file_sizes:
            print('\n Internal error: no output countTags file found.\n')
            sys.exit()


    def get_by_category(self, categ, mode=None):
        """ get counts by category """
        results = []
        ### By default, use with invoked mode (paired or single)
        mode = mode or self.mode
        ### get counts by category
        if categ in self.tags['counts']:
            for seq_id in self.tags['counts'][categ]:
                counts = self.tags['counts'][categ][seq_id][mode]
                results.append((seq_id, counts))
            return results
        else:
            errmsg = f" Category warning: {categ!r}, defined in {self.args.config!r}, not found in output countTags.\n"
            print(errmsg)
            return None


    def get_counts_by_sample(self, seq_type, mode=None):
        """ Function doc """
        results = [ {sample: []} for sample in self.samples ]
        ### By default, use with invoked mode (paired or single)
        mode = mode or self.mode
        ### if sequence category is mentioned (ex: 'Gender')
        if not seq_type in self.types:
            sys.exit("\n Error: sequence type '{}' not found among these: {}.\n".format(seq_type, self.types))
        for seq_id in self.tags['counts'][seq_type]:
            counts = self.tags['counts'][seq_type][seq_id][mode]
            for i, count in enumerate(counts):
                if seq_id[-4:] == '_rev':
                    pass
                else:
                    results[i].get(self.samples[i]).append(count)
        return results


    def get_by_sample(self, sample_name, seq_type=None, mode=None):
        """ Function doc """
        results = {}
        ### By default, use with invoked mode (paired or single)
        mode = mode or self.mode
        ### if not samples, return error and quit
        if not sample_name in self.samples:
            sys.exit("\n Error: sample '{}' not found among these: {}.\n".format(sample_name, self.samples))
        ### index of sample
        sample_index = self.samples.index(sample_name)
        ### if sequence type is mentioned (ex: 'sex')
        if seq_type:
            if not seq_type in self.types:
                sys.exit("\n Error: sequence type '{}' not found among these: {}.\n".format(seq_type, self.types))
            for seq_id in self.tags['counts'][seq_type]:
                for seq_id in self.tags['counts'][seq_type]:
                    results[seq_id] = self.tags['counts'][seq_type][seq_id][mode][sample_index]
        ### without sequence type
        else:
            for seq_type in self.tags['counts']:
                if seq_type not in results:
                    results[seq_type] = {}
                for seq_id in self.tags['counts'][seq_type]:
                    if seq_id not in results:
                        results[seq_type][seq_id] = {}
                    for seq_id in self.tags['counts'][seq_type]:
                        results[seq_type][seq_id] = self.tags['counts'][seq_type][seq_id][mode][sample_index]
        return results


    def to_file(self, output_dir='.', output_file='table.tsv', sep='\t'):
        """ write in tabuled file """
        rows = ""
        ### Add columns name at firt row
        rows += sep.join(['type', 'seq_id']) + sep + sep.join(self.samples) + '\n'
        ### Add data in rows
        for seq_type in self.tags['counts']:
            for seq_id in self.tags['counts'][seq_type]:
                counts = self.tags['counts'][seq_type][seq_id][self.mode]
                ### counts list to string
                counts = sep.join([str(count) for count in counts])
                rows += sep.join((seq_type, seq_id, counts)) + "\n"

        filepath = os.path.join(output_dir, output_file)
        if not os.path.isdir(output_dir):
            os.mkdir(output_dir)
        with open(filepath, 'w') as stream:
            stream.write(rows)
        return os.path.join(output_dir, output_file)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding:utf8 -*-

"""
From a bunch of files, determine the samples weither the files are single or paired.
"""

import os
import sys
import re

TAGS = [
    ('_R1_001', '_R2_001', 'illumina'),
    ('_R1.fastq', '_R2.fastq', 'illumina simplified'),
    ('_1.fastq', '_2.fastq', 'bio2m'),
    ('_1.fq', '_2.fq', 'other'),
    ('_1.tsv', '_2.tsv', 'countTags'),
]

def main():
    """ You can use this module stand alone. """
    #sample_list = find_paired(sys.argv[1:])
    class args:
        single = False
        yes = True
    sample_list = find(sys.argv[1:], args)
    print(*sample_list, sep="\n  ")


def find(files, args):
    """ Function doc """
    if args.single:
        return find_single(files)
    return find_paired(files, args)


def find_paired(files, args):
    """ Function doc """
    samples = []
    control = []
    for file in files:
        ### if file seems paired file (see TAGS)
        for tag in TAGS:
            if tag[0] in file:
                ### set mate file
                mate = file.replace(tag[0], tag[1])
                if os.path.isfile(mate):
                    sample = os.path.basename(re.sub('{}.*'.format(tag[0]), '', file))
                    samples.append((file, sample, '_1'))
                    samples.append((mate, sample, '_2'))
                    ### to control missing files
                    control += [file, mate]
    ### Control files
    f_not_found = []
    for file in files:
        if file not in control:
            f_not_found.append(file)
    if not control:
        msg = "\n No paired files found in {}, you may want to use '--single' argument.\n" \
                .format(os.path.dirname(files[0]))
        sys.exit(msg)
    if f_not_found:
        msg = "\n Be carefull, one or more file seems not be paired."
        msg += "\n No mate found for:"
        print(msg, *f_not_found, sep='\n\t')
        if not args.yes:
            valid = input("\n Would you like to continue? (these files will not be processed) [Y,n] ")
            if valid.lower() == 'n':
                sys.exit("\n Program aborted by user")
        else:
            if not args.yes:
                print("\n Program continues")
    return samples


def find_single(files):
    """ Function doc """
    samples = []
    for file in files:
        sample = os.path.basename(file)
        sample, ext = os.path.splitext(sample)
        if ext == '.gz':
            sample, ext = os.path.splitext(sample)
        if sample in samples:
            sys.exit("Maybe two files in the directory, with one compressed ({})".format(file))
        samples.append((file, sample))
    return samples


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding:utf8 -*-

"""
Check if input files are valid fastq or output countTags files.
- fastq, fastq.gz
- tsv, tsv.gz
"""


import os
import sys
import gzip


def main():
    """ You can use this module stand alone. """
    file = File(sys.argv[1])
    if file.is_valid:
        print("{} is valid {} file".format(sys.argv[1], file.type))
    else:
        print(file.errmsg)


class File:
    """ Class doc """

    def __init__(self, file):
        """ Class initialiser """
        self.file = file
        self.errmsg = ""
        self.is_valid, self.type = self._is_valid()


    def _is_valid(self):
        """ Function doc """
        ### Test if file is here
        if not os.path.isfile(self.file):
            if os.path.isdir(self.file):
                self.errmsg = "Error: Is a directory: {0}".format(self.file)
            else:
                self.errmsg = "Error: No such file: {0}".format(self.file)
            return False, None

        ### Test file type and consistency
        ftl = []  # First Three Lines
        ### open fastq or fastq.gz or countTags file
        try:
            with gzip.open(self.file, 'rt') as stream:
                for _ in range(3):
                    ftl.append(stream.readline())
        except OSError:
            with open(self.file) as stream:
                for _ in range(3):
                    ftl.append(stream.readline())
        ### check content of file
        if not self.errmsg:
            try:
                ### Check if fastq file
                if ftl[0][0] == "@" and ftl[1][0] in 'ATCGNatcgn' and ftl[2][0] == "+":
                    return True, 'fastq'
                ### Check if countTags file
                if ftl[0].split()[0] == 'tag' and ftl[1][0] in 'ATCGatcg' and ftl[2][0] in 'ATCGatcg':
                    if ftl[0].split()[1] == 'tag_names':
                        return True, 'countTags'
                    else:
                        self.errmsg = "\n Error, '{}' is valid countTags file, but without 'tag_names' field\n" \
                                      " run countTags with '--tag-names' option\n".format(self.file)
            except IndexError:
                self.errmsg = "Error: too small file: {}".format(self.file)
            else:
                self.errmsg = self.errmsg or "Error: {} does not seem a valid fastq or countTags file.".format(self.file)
        return False, None

if __name__ == "__main__":
    main()



### Const

import os
import sys
import shutil

# ~ APPPATH = os.path.dirname(os.path.realpath(__file__))
# ~ DEFAULT_TAGSET = "human-quality"


# ~ def get_tags_file(args):
    # ~ """ Function doc """
    # ~ ### Find tags file
    # ~ ## --tags used to specify tags file
    # ~ if args.tags:
        # ~ tags_file = args.tags
    # ~ ## tags file specified from --builtin-tags option (default: human)
    # ~ else:
        # ~ tags_file = BUILTIN_TAGS[args.builtin_tags]
    # ~ ### Is tag file here ?
    # ~ if not os.path.isfile(tags_file):
        # ~ print("\n FileError: tags file '{}' not Found.\n".format(tags_file))
        # ~ sys.exit(exit_gracefully(args))
    # ~ return tags_file


def exit_gracefully(args, files_type=None):
    """ Function doc """
    ### In debug mode, do not remove temporary files.
    if args.debug:
        print("\n In debug mode, you should remove manually {} temp directory".format(args.tmp_dir))
    elif args.tmp_dir:
            shutil.rmtree(args.tmp_dir)
    ### With EOF, the program exit succesfully.
    if  args.debug: print("EOF")



#!/usr/bin/env python3

""" Module Doc

input:
  Mandatory
    - fastq or output countTags files
  Optionnal
    - tags file
    - yaml file
output:
  - counts.tsv
  - kmerexplor.html
  - in lib directory
    - echarts-en.min.js
    - scripts.js
    - styles.css
"""


import os
import sys
import multiprocessing
# from functools import partial       # to send multiple arguments with pool.starmap
import tempfile
import gzip
import glob
import subprocess
import yaml

import info
from common import *
import checkFile as cf
import samples
from counts import Counts
from mk_results import TSV, HTML
from options import usage


APPPATH = os.path.dirname(os.path.realpath(__file__))


def main():
    """ Handle keyboard interrupt commands and launch program """
    ### 1. Manage command line arguments
    args = usage()

    ### load config file
    config = config2dict(args)
    if not config:
        sys.exit(f"Config error: {fileset.config!r} seems to be empty")

    ### If "ctrl C" is set, quit after executing exit_gracefully function
    try:
        run(args, config)
    except KeyboardInterrupt:
        print(f"Process interrupted by user", file=sys.stderr)
        exit_gracefully(args)


def config2dict(args):
    """ Load config yaml file as dict """
    config_file = args.setfiles['config']
    try:
        with open(config_file) as stream:
            return yaml.safe_load(stream)
    except FileNotFoundError:
        sys.exit(f"Error: no such file or directory: {config_file!r}")
    except yaml.scanner.ScannerError as err:
        sys.exit(f"Error {err}")


def run(args, config):
    """ Function doc """
    nprocs, files = args.cores, args.files
    if args.debug: print("Arguments:", args)

    ### 1. Define some PATH
    # create directory for temporay files
    args.tmp_dir = tempfile.mkdtemp(prefix=info.APPNAME.lower() + "-", dir=args.tmp_dir)
    # define countTags PATH
    if args.keep_counts:
        ## If the user wants to keep the countTag counts
        args.counts_dir = os.path.join(args.output, 'countTags')
        os.makedirs(args.counts_dir, exist_ok=True)
    else:
        ## Else counts will be removed at the exit_gracefully() function
        args.counts_dir = args.tmp_dir

    ### 2. Test validity files (Multiprocessed)
    with multiprocessing.Pool(processes=nprocs) as pool:
        # results = pool.map_async(partial(is_valid, args),files)        # Alternative
        results = pool.starmap_async(is_valid, [(file, args) for file in files])
        results.wait()
        files_in_error = []
        valid_types = set()
        ### Manage results
        for type,errmsg in results.get():
            if type:
                valid_types.add(type)
            else:
                files_in_error.append(errmsg)
        ### If a file(s) is not valid, exit with error message.
        if files_in_error:
            print(*files_in_error, sep='\n', file=sys.stderr)
            sys.exit(exit_gracefully(args))
        ### If a mix of valid file type are found, exit
        if len(valid_types) != 1:
            print("Multiples valid type found ({}).".format(*valid_types), file=sys.stderr)
        files_type = next(iter(valid_types))

    ### 3. If fastq files, determine paired (if '--paired' argument is specified)
    if files_type == 'fastq':
        sample_list = set_sample_list(files, args, files_type)
        if not sample_list:
            print("\n Error: no samples {} found\n".format('single' if args.single else 'paired'), file=sys.stderr)
            sys.exit(exit_gracefully(args, files_type))


    ### 4. If input files are fastq, run countTags (Multiprocessed)
    if files_type == 'fastq':
        ### Compute countTags with multi processing (use --core to specify cores counts)
        sys.stdout.write("\n ✨✨ Starting countTags, please wait.\n\n")
        with multiprocessing.Pool(processes=nprocs) as pool:
            results = pool.starmap_async(do_countTags, [(sample, args) for sample in sample_list])
            results.wait()
        samples_path = [f for f in glob.glob("{}/*.tsv".format(args.counts_dir))]
    else:
        samples_path = args.files

    ### 5. merge countTags tables
    sys.stdout.write("\n ✨✨ Starting merge of counts.\n")
    samples_path.sort()
    counts = Counts(samples_path, args)

    ### 6. Build results as html pages and tsv table
    sys.stdout.write("\n ✨✨ Build output html page.\n")
    table = TSV(counts, args)                                   # create TSV file
    charts = HTML(counts, args, info, config, args.setfiles['desc'])     # create results in html format

    ### 7. show results
    show_res(args,counts, table.tsvfile, charts.htmlfile, files_type)

    ### 8. exit gracefully program
    exit_gracefully(args, files_type)



def is_valid(file, args):
    """ Function doc """
    f = cf.File(file)
    if args.debug and f.is_valid:
        print("{} is valid {} file.".format(file, f.type))
    return f.type, f.errmsg


def set_sample_list(files, args, files_type):
    """ Function doc """
    if files_type == 'fastq':
        return samples.find(files, args)


def do_countTags(sample, args):
    """ Compute countTags """
    tags_file = args.setfiles['tags']
    countTag_cmd = '{0}/countTags --stranded --alltags --normalize --tag-names --merge-counts \
                    -b --merge-counts-colname {4} -k {1} \
                    --summary {5}/{4}.summary \
                    -i {2} {3} > {5}/{4}.tsv'.format(APPPATH, args.kmer_size, tags_file, sample[0],
                                                    ''.join(sample[1:]), args.counts_dir)
    if args.debug: print(f"{sample[1]}: Start countTags processing.\n{countTag_cmd}")
    os.popen(countTag_cmd).read()
    print("  {}: countTags processed.".format(''.join(sample[1:])))


def show_res(args, tags, tsvfile, htmlfile, files_type):
    """ output """
    print("\n ✨✨ Work is done, show:\n\n {}\n {}".format(os.path.abspath(tsvfile),
                                                           os.path.abspath(htmlfile)))
    if args.keep_counts and files_type == 'fastq' : print(" {}/".format(os.path.abspath(args.counts_dir)))
    print("\n")
    ### Launch results in a browser if possible
    try:
        subprocess.Popen(['x-www-browser', htmlfile])
    except:
        pass


if __name__ == "__main__":
    main()
